package ch.psi.daq.queryrest.controller;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import javax.annotation.Resource;
import javax.servlet.http.HttpServletResponse;
import javax.validation.Valid;

import org.apache.commons.lang3.tuple.Pair;
import org.apache.commons.lang3.tuple.Triple;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.context.ApplicationContext;
import org.springframework.http.MediaType;
import org.springframework.validation.Validator;
import org.springframework.web.bind.WebDataBinder;
import org.springframework.web.bind.annotation.InitBinder;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.ResponseBody;
import org.springframework.web.bind.annotation.RestController;

import ch.psi.daq.cassandra.config.CassandraConfig;
import ch.psi.daq.cassandra.request.validate.RequestProviderValidator;
import ch.psi.daq.common.json.deserialize.AttributeBasedDeserializer;
import ch.psi.daq.common.ordering.Ordering;
import ch.psi.daq.domain.DataEvent;
import ch.psi.daq.domain.json.ChannelName;
import ch.psi.daq.domain.reader.Backend;
import ch.psi.daq.query.analyzer.QueryAnalyzer;
import ch.psi.daq.query.config.QueryConfig;
import ch.psi.daq.query.model.Aggregation;
import ch.psi.daq.query.model.AggregationType;
import ch.psi.daq.query.model.Query;
import ch.psi.daq.query.model.QueryField;
import ch.psi.daq.query.model.ResponseFormat;
import ch.psi.daq.query.model.impl.BackendQuery;
import ch.psi.daq.query.model.impl.DAQQueries;
import ch.psi.daq.query.model.impl.DAQQuery;
import ch.psi.daq.query.model.impl.DAQQueryElement;
import ch.psi.daq.query.processor.ChannelNameCache;
import ch.psi.daq.query.processor.QueryProcessor;
import ch.psi.daq.query.request.ChannelsRequest;
import ch.psi.daq.query.request.ChannelsResponse;
import ch.psi.daq.queryrest.response.csv.CSVResponseStreamWriter;
import ch.psi.daq.queryrest.response.json.JSONResponseStreamWriter;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.common.collect.Lists;

@RestController
public class QueryRestController {

   private static final Logger LOGGER = LoggerFactory.getLogger(QueryRestController.class);

   public static final String CHANNELS = "/channels";
   public static final String QUERY = "/query";
   public static final String QUERIES = "/queries";

   @Resource
   private Validator queryValidator;
   private Validator requestProviderValidator = new RequestProviderValidator();

   @Resource
   private JSONResponseStreamWriter jsonResponseStreamWriter;

   @Resource
   private CSVResponseStreamWriter csvResponseStreamWriter;

   @Resource
   private ApplicationContext appContext;

   @Resource
   private ObjectMapper objectMapper;

   @Resource
   private Function<Query, QueryAnalyzer> queryAnalizerFactory;

   private Map<Backend, QueryProcessor> queryProcessors = new LinkedHashMap<>();
   private ChannelNameCache channelNameCache;

   @PostConstruct
   public void afterPropertiesSet() {
      List<Exception> exceptions = new ArrayList<>();
      
      try {
         QueryProcessor queryProcessor =
               appContext.getBean(QueryConfig.BEAN_NAME_CASSANDRA_QUERY_PROCESSOR, QueryProcessor.class);
         queryProcessors.put(queryProcessor.getBackend(), queryProcessor);
      } catch (Exception e) {
         exceptions.add(e);
         LOGGER.warn("");
         LOGGER.warn("##########");
         LOGGER.warn("Could not load query processor for cassandra.");
         LOGGER.warn("##########");
         LOGGER.warn("");
      }

      try {
         QueryProcessor queryProcessor =
               appContext.getBean(QueryConfig.BEAN_NAME_ARCHIVER_APPLIANCE_QUERY_PROCESSOR, QueryProcessor.class);
         queryProcessors.put(queryProcessor.getBackend(), queryProcessor);
      } catch (Exception e) {
         exceptions.add(e);
         LOGGER.warn("");
         LOGGER.warn("##########");
         LOGGER.warn("Could not load query processor for archiverappliance.");
         LOGGER.warn("##########");
         LOGGER.warn("");
      }
      
      if(queryProcessors.isEmpty()){
         LOGGER.error("No query processor could be loaded! Exceptions were: ");
         for (Exception exception : exceptions) {
            LOGGER.error("", exception);
         }
         
         throw new RuntimeException("No Backends available!");
      }

      channelNameCache =
            new ChannelNameCache(queryProcessors, appContext.getBean(CassandraConfig.BEAN_NAME_READ_TIMEOUT,
                  Integer.class).longValue());
   }

   @PreDestroy
   public void destroy() {
      channelNameCache.destroy();
   }

   @InitBinder
   protected void initBinder(WebDataBinder binder) {
      if (binder.getTarget() != null) {
         if (requestProviderValidator.supports(binder.getTarget().getClass())) {
            binder.addValidators(requestProviderValidator);
         }
         if (queryValidator.supports(binder.getTarget().getClass())) {
            binder.addValidators(queryValidator);
         }
      }
   }

   @RequestMapping(value = CHANNELS, method = {RequestMethod.GET, RequestMethod.POST},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<ChannelsResponse> getChannels(@RequestBody(required = false) ChannelsRequest request)
         throws Throwable {
      // in case not specified use defaults (e.g. GET)
      if (request == null) {
         request = new ChannelsRequest();
      }

      return channelNameCache.getChannels(request);
   }

   /**
    * Query specific channel names, and return only those.
    * 
    * @param channelName part of (or full) channel name
    * @return Collection of channel names matching the specified input channel name
    * @throws Throwable in case something goes wrong
    */
   @RequestMapping(value = CHANNELS + "/{channelName}", method = {RequestMethod.GET},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody Collection<ChannelsResponse> getChannels(@PathVariable(value = "channelName") String channelName)
         throws Throwable {
      return getChannels(new ChannelsRequest(channelName));
   }

   /**
    * Accepts the properties for the {@link DAQQuery} instance as stringified JSON query string.
    * 
    * @param jsonBody The {@link DAQQuery} properties sent as a JSON string, i.e. this is the
    *        stringified body of the POST request method
    * @param res the current {@link HttpServletResponse} instance
    * @throws Exception if reading the JSON string fails or if the subsequent call to
    *         {@link #executeQuery(DAQQuery, HttpServletResponse)} fails
    */
   @RequestMapping(
         value = QUERY,
         method = RequestMethod.GET)
   public void executeQueryBodyAsString(@RequestParam String jsonBody, HttpServletResponse res) throws Exception {
      DAQQuery query = objectMapper.readValue(jsonBody, DAQQuery.class);
      executeQuery(query, res);
   }

   /**
    * Executes a single query.
    * 
    * @param query the {@link DAQQuery}
    * @param res the {@link HttpServletResponse} instance associated with this request
    * @throws IOException thrown if writing to the output stream fails
    */
   @RequestMapping(
         value = QUERY,
         method = RequestMethod.POST,
         consumes = {MediaType.APPLICATION_JSON_VALUE})
   public void executeQuery(@RequestBody @Valid DAQQuery query, HttpServletResponse res) throws Exception {
      executeQueries(new DAQQueries(query), res);
   }

   /**
    * Accepts the properties for the {@link DAQQueries} instance as stringified JSON query string.
    * 
    * @param jsonBody The {@link DAQQueries} properties sent as a JSON string, i.e. this is the
    *        stringified body of the POST request method
    * @param res the current {@link HttpServletResponse} instance
    * @throws Exception if reading the JSON string fails or if the subsequent call to
    *         {@link #executeQueries(DAQQueries, HttpServletResponse)} fails
    */
   @RequestMapping(
         value = QUERIES,
         method = RequestMethod.GET)
   public void executeQueriesBodyAsString(@RequestParam String jsonBody, HttpServletResponse res) throws Exception {
      DAQQueries queries = objectMapper.readValue(jsonBody, DAQQueries.class);
      executeQueries(queries, res);
   }

   /**
    * Catch-all query method for getting data from the backend for both JSON and CSV requests.
    * <p>
    * The {@link DAQQueries} object will contain the concrete subclass based on the combination of
    * fields defined in the user's query. The {@link AttributeBasedDeserializer} decides which class
    * to deserialize the information into and has been configured (see
    * QueryRestConfig#afterPropertiesSet) accordingly.
    * 
    * @param queries the {@link DAQQueryElement}s
    * @param res the {@link HttpServletResponse} instance associated with this request
    * @throws IOException thrown if writing to the output stream fails
    */
   @RequestMapping(
         value = QUERIES,
         method = RequestMethod.POST,
         consumes = {MediaType.APPLICATION_JSON_VALUE})
   public void executeQueries(@RequestBody @Valid DAQQueries queries, HttpServletResponse res) throws Exception {
      try {
         LOGGER.debug("Executing queries '{}'", queries);

         if (ResponseFormat.JSON.equals(queries.getResponseFormat())) {
            // write the response back to the client using java 8 streams
            jsonResponseStreamWriter.respond(executeQueries(queries), queries, res);
         } else if (ResponseFormat.CSV.equals(queries.getResponseFormat())) {
            // it's a CSV request
            executeQueriesCsv(queries, res);
         } else {
            String message = String.format("Unsupported response format '%s'", queries.getResponseFormat().name());
            LOGGER.error(message);
            throw new RuntimeException(message);
         }

      } catch (Exception e) {
         LOGGER.error("Failed to execute query '{}'.", queries, e);
         throw e;
      }
   }

   /**
    * Returns data in CSV format to the client.
    */
   private void executeQueriesCsv(DAQQueries queries, HttpServletResponse res) throws Exception {

      for (DAQQueryElement query : queries) {
         if (!(query.getAggregationType() == null || AggregationType.value.equals(query.getAggregationType()))) {
            // We allow only no aggregation or value aggregation as
            // extrema: nested structure and not clear how to map it to one line
            // index: value is an array of Statistics whose size is not clear at initialization time
            String message = "CSV export does not support '" + query.getAggregationType() + "'";
            LOGGER.warn(message);
            throw new IllegalArgumentException(message);
         }
      }

      try {
         LOGGER.debug("Executing query '{}'", queries);
         // write the response back to the client using java 8 streams
         csvResponseStreamWriter.respond(executeQueries(queries), queries, res);
      } catch (Exception e) {
         LOGGER.error("Failed to execute query '{}'.", queries, e);
         throw e;
      }

   }

   public List<Entry<DAQQueryElement, Stream<Triple<BackendQuery, ChannelName, ?>>>> executeQueries(DAQQueries queries) {
      // set backends if not defined yet
      channelNameCache.setBackends(queries);

      List<Entry<DAQQueryElement, Stream<Triple<BackendQuery, ChannelName, ?>>>> results =
            new ArrayList<>(queries.getQueries().size());

      for (DAQQueryElement queryElement : queries) {
         Stream<Triple<BackendQuery, ChannelName, ?>> resultStreams =
               queryElement
                     .getBackendQueries()
                     .stream()
                     .filter(query -> {
                        QueryProcessor processor = queryProcessors.get(query.getBackend());
                        if (processor != null) {
                           return true;
                        } else {
                           LOGGER.warn("There is no QueryProcessor available for '{}'", query.getBackend());
                           return false;
                        }
                     })
                     .flatMap(query -> {
                        QueryProcessor processor = queryProcessors.get(query.getBackend());
                        QueryAnalyzer queryAnalizer = queryAnalizerFactory.apply(query);

                        // all the magic happens here
                           Stream<Entry<ChannelName, Stream<? extends DataEvent>>> channelToDataEvents =
                                 processor.process(queryAnalizer);
                           // do post-process
                           Stream<Entry<ChannelName, ?>> channelToData = queryAnalizer.postProcess(channelToDataEvents);

                           return channelToData.map(entry -> {
                              return Triple.of(query, entry.getKey(), entry.getValue());
                           });
                        });

         // Now we have a stream that loads elements sequential BackendQuery by BackendQuery.
         // By materializing the outer Stream the elements of all BackendQuery are loaded async
         // (speeds things up but requires also more memory - i.e. it relies on Backends not loading
         // all elements into memory at once)
         resultStreams = resultStreams.collect(Collectors.toList()).stream();

         results.add(Pair.of(queryElement, resultStreams));
      }

      return results;
   }

   /**
    * Returns the current list of {@link Ordering}s available.
    * 
    * @return list of {@link Ordering}s as String array
    */
   @RequestMapping(value = "ordering", method = {RequestMethod.GET}, produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<Ordering> getOrderingValues() {
      return Lists.newArrayList(Ordering.values());
   }

   /**
    * Returns the current list of {@link QueryField}s available.
    * 
    * @return list of {@link QueryField}s as String array
    */
   @RequestMapping(value = "queryfields", method = {RequestMethod.GET}, produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<QueryField> getQueryFieldValues() {
      return Lists.newArrayList(QueryField.values());
   }

   /**
    * Returns the current list of {@link Aggregation}s available.
    * 
    * @return list of {@link Aggregation}s as String array
    */
   @RequestMapping(value = "aggregations", method = {RequestMethod.GET}, produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<Aggregation> getAggregationsValues() {
      return Lists.newArrayList(Aggregation.values());
   }

   /**
    * Returns the current list of {@link AggregationType}s available.
    * 
    * @return list of {@link AggregationType}s as String array
    */
   @RequestMapping(value = "aggregationtypes", method = {RequestMethod.GET},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<AggregationType> getAggregationTypesValues() {
      return Lists.newArrayList(AggregationType.values());
   }

   /**
    * Returns the current list of {@link Backend} values.
    * 
    * @return list of {@link Backend}s as String array
    */
   @RequestMapping(value = "backends", method = {RequestMethod.GET}, produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<Backend> getDBModeValues() {
      return Lists.newArrayList(Backend.values());
   }
}
