package ch.psi.daq.queryrest.response.csv;

import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.Function;
import java.util.stream.Stream;

import javax.annotation.Resource;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.lang3.tuple.Triple;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.supercsv.cellprocessor.constraint.NotNull;
import org.supercsv.cellprocessor.ift.CellProcessor;
import org.supercsv.io.dozer.CsvDozerBeanWriter;
import org.supercsv.io.dozer.ICsvDozerBeanWriter;
import org.supercsv.prefs.CsvPreference;

import com.fasterxml.jackson.core.JsonEncoding;

import ch.psi.daq.domain.DataEvent;
import ch.psi.daq.domain.json.ChannelName;
import ch.psi.daq.query.analyzer.QueryAnalyzer;
import ch.psi.daq.query.model.Aggregation;
import ch.psi.daq.query.model.Query;
import ch.psi.daq.query.model.QueryField;
import ch.psi.daq.query.model.ResponseOptions;
import ch.psi.daq.query.model.impl.BackendQuery;
import ch.psi.daq.query.model.impl.DAQQueryElement;
import ch.psi.daq.queryrest.response.AbstractResponseStreamWriter;

/**
 * Takes a Java 8 stream and writes it to the output stream provided by the {@link ServletResponse}
 * of the current request.
 */
public class CSVResponseStreamWriter extends AbstractResponseStreamWriter {

   private static final char DELIMITER_CVS = ';';
   private static final char DELIMITER_ARRAY = ',';

   private static final Logger LOGGER = LoggerFactory.getLogger(CSVResponseStreamWriter.class);

   @Resource
   private Function<Query, QueryAnalyzer> queryAnalizerFactory;

   @Override
   public void respond(List<Entry<DAQQueryElement, Stream<Triple<BackendQuery, ChannelName, ?>>>> results,
         ResponseOptions options,
         HttpServletResponse response) throws Exception {
      response.setCharacterEncoding(JsonEncoding.UTF8.getJavaName());
      response.setContentType(CONTENT_TYPE_CSV);

      respondInternal(results, options, response);
   }

   private void respondInternal(List<Entry<DAQQueryElement, Stream<Triple<BackendQuery, ChannelName, ?>>>> results,
         ResponseOptions options, HttpServletResponse response) throws Exception {
      AtomicReference<Exception> exception = new AtomicReference<>();
      OutputStream out = handleCompressionAndResponseHeaders(options, response, CONTENT_TYPE_CSV);
      List<ICsvDozerBeanWriter> beanWriters = new ArrayList<>();

      results.forEach(entry -> {
         DAQQueryElement query = entry.getKey();

         Set<QueryField> queryFields = query.getFields();
         List<Aggregation> aggregations = query.getAggregations();
         int aggregationsSize = (aggregations != null ? aggregations.size() : 0);

         Set<String> fieldMapping = new LinkedHashSet<>(queryFields.size() + aggregationsSize);
         List<String> header = new ArrayList<>(queryFields.size() + aggregationsSize);

         AtomicReference<CellProcessor[]> processorsRef = new AtomicReference<>();

         entry.getValue()
               .sequential()
               .forEach(triple -> {
                  try {
                     CellProcessor[] processors = processorsRef.get();
                     ICsvDozerBeanWriter beanWriter;

                     if (processors == null) {
                        processors = setupCellProcessors(query, triple.getLeft(), fieldMapping, header);
                        processorsRef.set(processors);

                        beanWriter = setupBeanWriter(fieldMapping, out);
                        beanWriters.add(beanWriter);
                     } else {
                        beanWriter = beanWriters.get(beanWriters.size() - 1);
                     }

                     writeToOutput(triple, processors, beanWriter, header);
                  } catch (Exception e) {
                     LOGGER.warn("Could not write CSV of '{}'", triple.getMiddle(), e);
                     exception.compareAndSet(null, e);
                  }
               });

         if (!beanWriters.isEmpty()) {
            try {
               beanWriters.get(beanWriters.size() - 1).flush();
            } catch (Exception e) {
               LOGGER.error("Could not flush ICsvDozerBeanWriter.", e);
               exception.compareAndSet(null, e);
            }
         }
      });

      for (ICsvDozerBeanWriter beanWriter : beanWriters) {
         try {
            beanWriter.close();
         } catch (Exception e) {
            LOGGER.error("Could not close ICsvDozerBeanWriter.", e);
            exception.compareAndSet(null, e);
         }
      }

      if (exception.get() != null) {
         throw exception.get();
      }
   }

   /**
    * Sets up the bean writer instance.
    * 
    * @throws Exception
    * 
    */
   private ICsvDozerBeanWriter setupBeanWriter(Set<String> fieldMapping, OutputStream out) throws Exception {
      CsvPreference preference = new CsvPreference.Builder(
            (char) CsvPreference.EXCEL_NORTH_EUROPE_PREFERENCE.getQuoteChar(),
            DELIMITER_CVS,
            CsvPreference.EXCEL_NORTH_EUROPE_PREFERENCE.getEndOfLineSymbols()).build();

      ICsvDozerBeanWriter beanWriter = new CsvDozerBeanWriter(new OutputStreamWriter(out), preference);
      // configure the mapping from the fields to the CSV columns
      beanWriter.configureBeanMapping(DataEvent.class, fieldMapping.toArray(new String[fieldMapping.size()]));
      return beanWriter;
   }

   /**
    * Sets up the array of {@link CellProcessor}s needed for later configuration of the bean writer.
    * 
    * Cell processors are an integral part of reading and writing with Super CSV - they automate the
    * data type conversions, and enforce constraints. They implement the chain of responsibility
    * design pattern - each processor has a single, well-defined purpose and can be chained together
    * with other processors to fully automate all of the required conversions and constraint
    * validation for a single CSV column.
    * 
    * @param daqQuery The current {@link DAQQueryElement}
    * @param backendQuery One BackendQuery of the current {@link DAQQueryElement}
    * @return Array of {@link CellProcessor} entries
    */
   private CellProcessor[] setupCellProcessors(DAQQueryElement daqQuery, BackendQuery backendQuery,
         Set<String> fieldMapping, List<String> header) {
      Set<QueryField> queryFields = daqQuery.getFields();
      List<Aggregation> aggregations = daqQuery.getAggregations();

      List<CellProcessor> processorSet =
            new ArrayList<>(queryFields.size() + (aggregations != null ? aggregations.size() : 0));

      boolean isNewField;
      QueryAnalyzer queryAnalyzer = queryAnalizerFactory.apply(backendQuery);

      for (QueryField field : queryFields) {
         if (!(QueryField.value.equals(field) && queryAnalyzer.isAggregationEnabled())) {
            isNewField = fieldMapping.add(field.name());

            if (isNewField) {
               header.add(field.name());
               processorSet.add(new ArrayProcessor(new NotNull(), DELIMITER_ARRAY));
            }
         }
      }

      if (aggregations != null && queryAnalyzer.isAggregationEnabled()) {
         for (Aggregation aggregation : daqQuery.getAggregations()) {
            isNewField = fieldMapping.add("value." + aggregation.name());

            if (isNewField) {
               header.add(aggregation.name());
               processorSet.add(new ArrayProcessor(new NotNull(), DELIMITER_ARRAY));
            }
         }
      }
      return processorSet.toArray(new CellProcessor[processorSet.size()]);
   }

   @SuppressWarnings("unchecked")
   private void writeToOutput(Triple<BackendQuery, ChannelName, ?> triple, CellProcessor[] processors,
         ICsvDozerBeanWriter beanWriter, List<String> header) throws IOException {
      if (triple.getRight() instanceof Stream) {
         beanWriter.writeComment("");
         beanWriter.writeComment("Start of " + triple.getMiddle());
         beanWriter.writeHeader(header.toArray(new String[header.size()]));

         Stream<DataEvent> eventStream = (Stream<DataEvent>) triple.getRight();
         eventStream
               .forEach(
               event -> {
                  try {
                     beanWriter.write(event, processors);
                  } catch (Exception e) {
                     LOGGER.error("Could not write elements for '{}-{}'", event.getChannel(),
                           event.getPulseId(), e);
                  }
               });
      } else {
         String message = "Type '" + triple.getRight().getClass() + "' not supported.";
         LOGGER.error(message);
      }
   }

}
